/*
 * ocache_low.c --- low I/O and filename procedures
 *      by k-chinen@is.aist-nara.ac.jp, 2000-2002.
 *
 *  this routines is splitted from xsk.c of kotetu-1.2.
 *
 * $Id$
 */


pthread_mutex_t sc_lock = PTHREAD_MUTEX_INITIALIZER;

#define DF(x,n) volatile long long sc_##x=0; char *sc_msg_##x=n
#define CX(x)   MUTEX_LOCK(&stat_lock); sc_##x++; MUTEX_UNLOCK(&stat_lock)
#define SH(p,x) fprintf(p, "%-32s %32lld\n",\
                    sc_msg_##x, (long long)sc_##x)

DF(iread,   "index  read");
DF(pread,   "packed read");
DF(ireadv,  "index  readv");
DF(preadv,  "packed readv");
DF(iwrite,  "index  write");
DF(pwrite,  "packed write");
DF(iwritev, "index  writev");
DF(pwritev, "packed writev");
DF(iopen,   "index  open");
DF(popen,   "packed open");
DF(iclose,  "index  close");
DF(pclose,  "packed close");
DF(ilseek,  "index  lseek");
DF(plseek,  "packed lseek");
DF(junk,    "junk");


char *cache_dir_prefix = "/var/tmp/cache";
int radix_disk = -1;

char *cache_index_prefix = NULL;


/*
 * When @ (at-mark) is found in prefix of cache dirctory,
 * the program recognize using many disk and its pattern.
 */
void
check_disk_spliting()
{
    char *p;
    int v, i;
    char tmp[L_PATH];

    v = -1;
    p = cache_dir_prefix;
    while(*p) {
        if(*p=='@') {
            v = atoi(p+1);
            *p = '\0';
            break;
        }
        p++;
    }

    if(v>0) {
        radix_disk = v;

        Trace(" Spricing: cache directories stored into %d disks\n",
            radix_disk);

#if 0
        for(i=0;i<radix_disk;i++) {
            Trace("   disk# %3d: %s%d\n", i, cache_dir_prefix, i);
        }
#endif
    }

#if 0
    show_bk2fs_map();
    show_fs2bk_map();
#endif
}



/*****
 ***** FILENAME ASSIGNMENT
 *****/

/*
 * kotetu-1.2:
 *  /c0/0/0/INDEX
 *  /c0/0/0/BODY
 *
 * 1.5, 1.6 and leter:
 *  /c0/0/0-INDEX
 *  /c0/0/0-BODY
 *
 */

inline
char*
assign_dir_bynumber(char *name, char *prefix, int h1)
{
    if(radix_disk<=0) {
        sprintf(name, "%s/%d", prefix, h1);
    }
    else {
        sprintf(name, "%s%d/%d",
            prefix, bk2fs[h1], h1);
    }

    return name;
}

inline
char*
assign_dir_bynumber2(char *name, char *prefix, int h1, int h2)
{
    if(radix_disk<=0) {
        sprintf(name, "%s/%d/%d", prefix, h1, h2);
    }
    else {
        sprintf(name, "%s%d/%d/%d",
            prefix, bk2fs[h1], h1, h2);
    }

    return name;
}

inline
char*
assign_file_bynumber4(char *name,
    char *prefix, int h1, int h2, int h3, int h4, char *base)
{
    char tmp[L_PATH];

    assign_dir_bynumber2(name, prefix, h1, h2);
/*
    sprintf(tmp, "/%d-%s", h3*RADIX_H3_MAX+h4, base);
*/
    sprintf(tmp, "/%d-%s", FNAME(h3, h4), base);
    strcat(name, tmp);

    return name;
}
inline
char*
assign_file4(char *name, char *prefix, skcache_entry_t *e, char *base)
{
    if(e->h3<0) {
        Error("assign_file4: h1,h2,h3,h4 %d,%d,%d,%d", e->h1, e->h2, e->h3);
    }
    assign_file_bynumber4(name, prefix, e->h1, e->h2, e->h3, e->h4, base);

    return name;
}


inline
char*
assign_file_bynumber3(char *name,
    char *prefix, int h1, int h2, int h3, char *base)
{
    char tmp[L_PATH];

    assign_dir_bynumber(name, prefix, h1);
    sprintf(tmp, "/%d/%d-%s", h2, h3, base);
    strcat(name, tmp);

    return name;
}
inline
char*
assign_file3(char *name, char *prefix, skcache_entry_t *e, char *base)
{
    if(e->h3<0) {
        Error("assign_file3: h1,h2,h3 %d,%d,%d", e->h1, e->h2, e->h3);
    }
    assign_file_bynumber3(name, prefix, e->h1, e->h2, e->h3, base);

    return name;
}

inline
char*
assign_file_bynumber2(char *name, char *prefix, int h1, int h2, char *base)
{
    char tmp[L_PATH];

    assign_dir_bynumber(name, prefix, h1);
#ifdef NP
    sprintf(tmp, "/%d/%s", h2, base);
#else
    sprintf(tmp, "/%d-%s", h2, base);
#endif
    strcat(name, tmp);

    return name;
}
inline
char*
assign_file2(char *name, char *prefix, skcache_entry_t *e, char *base)
{
    assign_file_bynumber2(name, prefix, e->h1, e->h2, base);

    return name;
}


inline
char*
assign_indexfilename(char *name, skcache_entry_t *e)
{
    assign_file2(name, cache_index_prefix, e, INDEX_FILENAME);
    return name;
}

inline
char*
assign_bodyfilename(char *name, skcache_entry_t *e)
{
#ifdef NP
    assign_file4(name, cache_dir_prefix, e, BODY_FILENAME);
#else
    assign_file2(name, cache_dir_prefix, e, BODY_FILENAME);
#endif
    return name;
}


int
build_dir_bynumber(int d1, int d2)
{
    int i;
    int ret=0;
    int chk;
    char name[L_PATH];
    char tmp[L_PATH];

    assign_dir_bynumber(name, cache_dir_prefix, d1);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() index '%s' 1 (%d)", name, errno);
        ret = -1;
    }

#if 1
    assign_dir_bynumber(name, cache_index_prefix, d1);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() body '%s' 2 (%d)", name, errno);
        ret = -1;
    }
#endif

#ifdef NP
    assign_dir_bynumber2(name, cache_index_prefix, d1, d2);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() body '%s' 3 (%d)", name, errno);
        ret = -1;
    }
#endif

out:
    return ret;
}


int
rm_fr(char *base)
{
    char tmp[L_PATH];
    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    struct stat sbuf;
    int chk;

    if(base==NULL)
        return -1;

#if 0
Trace("rm_fr: called '%s'\n", base);
#endif

    chk = rmdir(base);
    if(chk) {
        if(errno==ENOTEMPTY) {

            dirp = opendir(base);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.' && dp->d_name[1]=='\0') {
                        continue;
                    }
                    if(dp->d_name[0]=='.'
                        && dp->d_name[1]=='.' && dp->d_name[2]=='\0') {
                        continue;
                    }
                    /*
                    strcpy(tmp, base);
                    strcat(tmp, "/");
                    strcat(tmp, dp->d_name);
                    */
                    pathncat(tmp, L_PATH, base, dp->d_name);

                    chk = stat(tmp, &sbuf);
                    if(chk) {
                        Error("rm_fr: fail stat() '%s' (%d)",
                            tmp, errno);
                        continue;
                    }

                    if(S_ISREG(sbuf.st_mode)) {
                        unlink(tmp);
                    }
                    if(S_ISDIR(sbuf.st_mode)) {
                        rm_fr(tmp);
                    }
                }
                closedir(dirp);
            }

            chk = rmdir(base);

            if(chk) {
                Error("rm_fr: fail rmdir() '%s' after sweep (%d)",
                    base, errno);
            }

        }
    }

    return 0;
}




/*****
 ***** fd caching 
 *****      map file to fd
 *****      position
 *****      compaction
 *****/

int fdcaching_flag=1;

pthread_mutex_t fd_assign_lock = PTHREAD_MUTEX_INITIALIZER;
int *ifd_map, *bfd_map;
int *ifd_ref, *bfd_ref;
int *ifd_mod, *bfd_mod;
off_t *ifd_pos, *bfd_pos;

int nfd_cached = 0;
int nfd_cached_hhwater = 1024, nfd_cached_hwater = 768, nfd_cached_lwater = 32;

int fd_comp_running = 0;
int fd_comp_width = 128;
int fd_comp_last = 0;
int ifd_keeptime = 4; /* greater (not include equal) */
int bfd_keeptime = 2; /* greater (not include equal) */


/*
 * return:
 *      0   right
 *      !0  strange
 */

off_t
raw_pos_verify(int fd, off_t pos, char *fname, int line)
{
    off_t apos;
    int c;

    if(fd<0) {
        Error("pv: fail fd# %d, pos %lld : %s %d",
            fd, (long long)pos, fname, line);
        return 0;
    }
    if(pos<(off_t)0) {
#if 0
        Error("pv: fail fd# %d, pos %lld : %s %d",
            fd, (long long)pos, fname, line);
#endif
        return 0;
    }

    c = 0;
try:
    c++;
    CX(junk);
    apos = lseek(fd, 0, SEEK_CUR);
    if(apos<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try;
        }
        Error("pv: fd# %d, error %d : %s %d",
            fd, errno, fname, line);
    }
    
    if(apos==pos) {
#if 0
        Trace("pv: fd# %d, pos view %lld, actual %lld ; equal\n",
            fd, (long long)pos, (long long)apos);
#endif
        return 0;
    }
    else {
        Error("pv: fd# %d, pos view %lld, actual %lld ; diff %lld : %s %d",
            fd, (long long)pos, (long long)apos, (long long) apos-pos,
            fname, line);
        return apos-pos;
    }
}


#ifdef TRACE_POS
#define POSCHECKF(fpos) \
    { if(fpos<0 && fpos>=dir_num) { \
        Error("chkF: fpos %d out of range [%d..%d) : %s %d", \
            fpos, 0, dir_num, __FILE__, __LINE__); \
      } \
    }
#define POSCHECKI(vpos) \
    { if(vpos<0L || vpos>(Onblk_file*ESIZE)) { \
        Error("chkI: index overrun, pos %lldB <%lds+%ld> : %s %d", \
            (long long)vpos, (long)(vpos/ESIZE), (vpos%ESIZE), \
            __FILE__, __LINE__); \
      } else { \
      } \
    }
#define POSCHECKP(vpos) \
    { if(vpos<0L || vpos>(Onblk_file*Oblksize)) { \
        Error("chkP: pack overrun, pos %lldB <%lds+%ld>: %s %d", \
            (long long)vpos, \
            (long)(vpos/Oblksize), (vpos%Oblksize), \
            __FILE__, __LINE__); \
      } else { \
      } \
    }

#else /* TRACE_POS */

#define POSCHECKF(x)        (void)1
#define POSCHECKI(x)        (void)1
#define POSCHECKP(x)        (void)1

#endif /* TRACE_POS */


#ifdef TRACE_POS2

#define POSVERIFY(xxf,xxp)  raw_pos_verify(xxf, xxp, __FILE__, __LINE__)
#define LENVERIFY(xxf)      (void)1

#else

#define POSVERIFY(xxf,xxp)  (void)1
#define LENVERIFY(xxf)      (void)1

#endif /* TRACE_POS2 */


extern time_t boot_date;

/*
 * seeking for packed file.
 */
inline
static
off_t 
Iseek(int fpos, int sfd, off_t pos)
{
    off_t dst;
    off_t dpos;
    off_t last_pos;
    int ddir;
    int c;

    /* NOTE:
        fpos    packed file number
        pos     file position of goal
        dst     file position of system returned
        ddir    direction for system-call applied
        dpos    position for system-call applied
     */

    POSCHECKF(fpos);
    POSCHECKI(pos);

    if(fpos<0 && fpos>=dir_num) {
        Error("Iseek: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    } 
    if(pos<(off_t)0 || pos>(off_t)(Onblk_file*ESIZE) ) {
        Error("Iseek: ignore position %lld [0..%lld] with dir# %d",
            (long long)pos, (long long)(Onblk_file*ESIZE), fpos);
        goto out;
    }

    last_pos = ifd_pos[fpos];

#ifndef NO_SEEKFAKE
    if(ifd_pos[fpos]<0)  {
            ddir = SEEK_SET;    dpos = pos;
    }
    else {
        if(pos == ifd_pos[fpos]) {
#ifdef TRACE_POS2
    Trace("Iseek: dir# %d, fd# %d, pos %lld, last %lld => no-move\n",
        fpos, sfd, (long long)pos, (long long)ifd_pos[fpos]);
#endif
            dst = ifd_pos[fpos];
            goto no_move;
        }
        if(pos==(off_t)0) {
            ddir = SEEK_SET;    dpos = 0;
        }
        else {              
            ddir = SEEK_CUR;    dpos = pos - ifd_pos[fpos];
        }
    }
#else
    {
            ifd_pos[fpos] = -1;
            ddir = SEEK_SET;    dpos = pos;
    }
#endif /* FAKE_SEEK */

#ifdef TRACE_POS2
    Trace("Iseek: start dir# %d, fd# %d, pos %d %lld, last %lld => lseek %lld %d\n",
        fpos, sfd, (long)(pos/ESIZE), (long long)pos,
        (long long)ifd_pos[fpos], (long long)dpos, ddir);
#endif

#if 0
    POSCHECKI(ifd_pos[fpos]);
#endif

    c = 0;
try_seek:
    c++;
    CX(ilseek);
    dst = lseek(sfd, dpos, ddir);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        Error(
"Iseek: fail dir# %d, fd# %d, cannot seek <%lldB dir %d> %lldB on index file : try %d errno %d",
            fpos, sfd, (long long)dpos, ddir, (long long)pos, c, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        dst = ifd_pos[fpos] = -1;
        goto out;
    }
    ifd_pos[fpos] = dst;
#if 0
    POSCHECKI(dst);
#endif

no_move:
out:
    POSCHECKI(dst);
#if 1
    if(dst<0L || dst>(Onblk_file*ESIZE)) { 
        Error("Iseek: dir# %d, fd# %d, goto %lld, last %lld / shift %lld dir %d / dst %lld",
            fpos, sfd,
            (long long)pos,
            (long long)last_pos,
            /*
            (long long)ifd_pos[fpos],
            */
            (long long)dpos, ddir,
            (long long)dst);
        ifd_pos[fpos] = -1L;
    }
#endif
    POSVERIFY(sfd, dst);

    return dst;
}


inline
static
off_t 
Pseek(int fpos, int sfd, off_t pos)
{
    off_t dst;
    off_t dpos;
    int ddir;
    int c;

    POSCHECKF(fpos);
    POSCHECKP(pos);

    if(fpos<0 && fpos>=dir_num) {
        Error("Pseek: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    if(pos<(off_t)0 || pos>(off_t)(Onblk_file*Oblksize)) {
        Error("Pseek: ignore position %lld [0..%lld] with dir# %d",
            (long long)pos, (long long)(Onblk_file*Oblksize), fpos);
    }

#ifndef NO_SEEKFAKE
    if(bfd_pos[fpos]<0) {  ddir = SEEK_SET;    dpos = pos; }
    else {
        if(pos == bfd_pos[fpos]) {
#ifdef TRACE_POS2
    Trace("Pseek: dir# %d, fd# %d, pos %lld, last %lld => no-move\n",
        fpos, sfd, (long long)pos, (long long)bfd_pos[fpos]);
#endif
            dst = pos;
            goto nomove;
        }
        if(pos==(off_t)0) { ddir = SEEK_SET;    dpos = 0; }
        else {              ddir = SEEK_CUR;    dpos = pos - bfd_pos[fpos]; }
    }
#else
    bfd_pos[fpos] = -1;
    ddir = SEEK_SET;    dpos = pos;
#endif

#ifdef TRACE_POS2
    Trace("Pseek: start dir# %d, fd# %d, goto pos %d %lld, last %lld => lseek %lld %d\n",
        fpos, sfd, (long)(pos/Oblksize), (long long)pos,
        (long long)bfd_pos[fpos], (long long)dpos, ddir);
#endif


    c = 0;
try_seek:
    c++;
    CX(plseek);
    dst = lseek(sfd, dpos, ddir);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        Error("Pseek: fail dir# %d, fd# %d, cannot seek <%lldB dir %d> %lldB on index file : try %d errno %d",
            fpos, sfd,
            (long long)dpos, ddir, (long long)pos, c, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);

        dst = bfd_pos[fpos] = -1;
        goto out;
    }
    bfd_pos[fpos] = dst;


nomove:
out:
    POSCHECKP(dst);
    POSVERIFY(sfd, dst);

    return dst;
}


int
Iread(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iread: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ifd_pos[fpos];
    c = 0;
try:
    c++;
    CX(iread);
    raw_ret = read(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Iread: fail read(); dir# %d, fd %d, pos %lld, len %d (%d)",
                fpos, fd, (long long)start_pos, len, errno);
            goto try;
        }
        Error("Iread: fail dir# %d, fd# %d, pos %lldB, len %d on index file : try %d errno %d",
            fpos, fd, (long long)start_pos, len, c, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %lld",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, (long long)ifd_pos[fpos]);
        ifd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        ifd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iread: dir# %d, fd# %d, read %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)ifd_pos[fpos]);
#endif
    }

    return raw_ret;
}

int
Ireadv(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Ireadv: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ifd_pos[fpos];
    c = 0;
try:
    c++;
    CX(ireadv);
    raw_ret = readv(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Ireadv: fail readv(); dir# %d, fd %d, pos %lld, vc %d (%d)",
                fpos, fd, (long long)start_pos, vc, errno);
            goto try;
        }
        Error("Ireadv: fail dir# %d, fd# %d, pos %lldB, vc %d on index file : try %d errno %d",
            fpos, fd, (long long)start_pos, vc, c, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %lld",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, (long long)ifd_pos[fpos]);
        ifd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        ifd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Ireadv: dir# %d, fd# %d, read %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)ifd_pos[fpos]);
#endif
    }

    return raw_ret;
}

int
Iwrite(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iwrite: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    POSCHECKI(ifd_pos[fpos]);

    start_pos = ifd_pos[fpos];
    c = 0;
try:
    c++;
    CX(iwrite);
    raw_ret = write(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Iwrite: fail write() index; dir# %d, fd %d, pos %lld, len %d (%d)",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        Error("Iwrite: fail write() index; dir# %d, fd %d, pos %lld, len %d (%d)",
            fpos, fd, (long long)start_pos, len, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        ifd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        ifd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iwrite: dir# %d, fd# %d, write %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)ifd_pos[fpos]);
#endif
        POSCHECKI(ifd_pos[fpos]);
#if 1
        if(ifd_pos[fpos]<0 || ifd_pos[fpos]>(Onblk_file*ESIZE)) { 
    Trace("Iwrite: dir# %d, fd# %d, buf %p, len %ld; write %ld moveto %lld\n",
                fpos, fd, buf, (long)len,
                (long)raw_ret, (long long)ifd_pos[fpos]);
        }
#endif
    }

    return raw_ret;
}

int
Iwritev(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Iwritev: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = ifd_pos[fpos];
    c = 0;
try:
    c++;
    CX(iwritev);
    raw_ret = writev(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Iwritev: fail writev() index; dir# %d, fd %d, pos %lld, vc %d (%d)",
                fpos, fd, start_pos, vc, errno);
            */
            goto try;
        }
        Error("Iwritev: fail writev() index; dir# %d, fd %d, pos %lld, vd %d (%d)",
            fpos, fd, (long long)start_pos, vc, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        ifd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        ifd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Iwritev: dir# %d, fd# %d, write %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)ifd_pos[fpos]);
#endif
        POSCHECKI(ifd_pos[fpos]);
    }

    return raw_ret;
}


int
Pread(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pread: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bfd_pos[fpos];
    c = 0;
try:
    c++;
    CX(pread);
    raw_ret = read(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Pread: fail read() data; dir# %d, fd %d, pos %lld, len %d (%d)",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        Error("Pread: fail read() data; dir# %d, fd %d, pos %lld, len %d (%d)",
            fpos, fd, start_pos, len, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        bfd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        bfd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pread: dir# %d, fd# %d, read %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)bfd_pos[fpos]);
#endif
    }

    return raw_ret;
}

int
Preadv(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Preadv: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bfd_pos[fpos];
    c = 0;
try:
    c++;
    CX(preadv);
    raw_ret = readv(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Preadv: fail readv() data; dir# %d, fd %d, pos %lld, vc %d (%d)",
                fpos, fd, start_pos, vc, errno);
            */
            goto try;
        }
        Error("Preadv: fail readv() data; dir# %d, fd %d, pos %lld, vc %d (%d)",
            fpos, fd, (long long)start_pos, vc, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        bfd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        bfd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Preadv: dir# %d, fd# %d, read %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)bfd_pos[fpos]);
#endif
    }

    return raw_ret;
}


int
Pwrite(int fpos, int fd, void* buf, int len)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pwrite: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bfd_pos[fpos];
    c = 0;
try:
    c++;
    CX(pwrite);
    raw_ret = write(fd, buf, len);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            /*
            Error("Pwrite: fail write() data; dir# %d, fd %d, pos %lld, len %d (%d)",
                fpos, fd, start_pos, len, errno);
            */
            goto try;
        }
        Error("Pwrite: fail write() data; dir# %d, fd %d, pos %lld, len %d (%d)",
                fpos, fd, (long long)(long long)(long long)(long long)(long long)(long long)(long long)(long long)start_pos, len, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        bfd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        bfd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pwrite: dir# %d, fd# %d, write %ld/%ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long)len, (long long)bfd_pos[fpos]);
#endif
        POSCHECKP(bfd_pos[fpos]);
    }

    return raw_ret;
}

int
Pwritev(int fpos, int fd, struct iovec *vb, int vc)
{
    int ret;
    int raw_ret;
    int c;
    off_t start_pos;

    if(fpos<0 && fpos>=dir_num) {
        Error("Pwritev: fpos %d out of range [%d..%d)",
            fpos, 0, dir_num);
    }
    POSCHECKF(fpos);
    start_pos = bfd_pos[fpos];
    c = 0;
try:
    c++;
    CX(pwritev);
    raw_ret = writev(fd, vb, vc);
    if(raw_ret<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            Error("Pwritev: fail writev() data; dir# %d, fd %d, pos %lld, vc %d (%d)",
                fpos, fd, start_pos, vc, errno);
            goto try;
        }
        Error("Pwritev: fail writev() data; dir# %d, fd %d, pos %lld, vc %d (%d)",
            fpos, fd, (long long)start_pos, vc, errno);
        Error("  fpos# %d, map %d, ref %d, mod %d(%d), pos %d",
            fpos, ifd_map[fpos], ifd_ref[fpos],
            ifd_mod[fpos], ifd_mod[fpos]-boot_date, ifd_pos[fpos]);
        bfd_pos[fpos] = -1;
    }
    if(raw_ret>0) {
        bfd_pos[fpos] += raw_ret;
#ifdef TRACE_POS2
        Trace("Pwritev: dir# %d, fd# %d, write %ld, moveto %lld\n",
            fpos, fd, (long)raw_ret, (long long)bfd_pos[fpos]);
#endif
        POSCHECKP(bfd_pos[fpos]);
    }

    return raw_ret;
}



inline
static
int
ifd_raw_closeP(int fpos)
{
    int ret;
    int chk;
    char fk;
    int xrfd;

    ret = 0;
    xrfd = ifd_map[fpos];

    MUTEX_LOCK(&fd_assign_lock);

    fk = 'd';
    ifd_map[fpos] = -1;
    ifd_ref[fpos] = 0;
    ifd_pos[fpos] = 0;
    ifd_mod[fpos] = 0;
    
    nfd_cached--;

#ifdef TRACE_FD
    Trace("ifd_rawcP: pos %6d; fd# %5d %2d; use %5d %c\n",
        fpos, xrfd, ifd_ref[fpos], nfd_cached, fk);
#endif

    MUTEX_UNLOCK(&fd_assign_lock);

    CX(iclose);
    close(xrfd);

out:
    return ret;
}

inline
static
int
bfd_raw_closeP(int fpos)
{
    int ret;
    int chk;
    char fk;
    int xrfd;

    ret = 0;
    xrfd = bfd_map[fpos];

    MUTEX_LOCK(&fd_assign_lock);

    fk = 'd';
    bfd_map[fpos] = -1;
    bfd_ref[fpos] = 0;
    bfd_pos[fpos] = 0;
    bfd_mod[fpos] = 0;
    
    nfd_cached--;

#ifdef TRACE_FD
    Trace("bfd_rawcP: pos %6d; fd# %5d %2d; use %5d %c\n",
        fpos, xrfd, bfd_ref[fpos], nfd_cached, fk);
#endif

    MUTEX_UNLOCK(&fd_assign_lock);

    CX(pclose);
    close(xrfd);

out:
    return ret;
}


int
ifd_raw_close(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int fpos;
    int xrfd;

    ret = 0;

    xrfd = *rfd;
    if(*rfd<0) {
        goto out;
    }

    fpos = FDPOOL_POS(e->h1, e->h2);

    ret = ifd_raw_closeP(fpos);

#ifdef TRACE_FD
    Trace("ifd_raw_c: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d; use %5d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, xrfd, ifd_ref[fpos], nfd_cached);
#endif

out:
    return ret;
}

int
bfd_raw_close(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int fpos;
    int xrfd;

    ret = 0;

    xrfd = *rfd;
    if(*rfd<0) {
        goto out;
    }

    fpos = FDPOOL_POS(e->h1, e->h2);

    ret = bfd_raw_closeP(fpos);

#ifdef TRACE_FD
    Trace("bfd_raw_c: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d; use %5d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, xrfd, bfd_ref[fpos], nfd_cached);
#endif

out:
    return ret;
}


int
ifd_close(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    int xrfd;
    int chk;
    char fk;
    int now;

    ret = 0;

    xrfd = *rfd;
    if(*rfd<0) {
        goto out;
    }

    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = -1; 
    xe->h4 = -1; 

    fpos = FDPOOL_POS_E(xe);

    if(!fdcaching_flag) {
        ret = ifd_raw_close(rfd, bk, e);
        *rfd = -1;
        goto out;
    }

    MUTEX_LOCK(&fd_assign_lock);

    if(ifd_ref[fpos]>0) {
        ifd_ref[fpos]--;
        ifd_mod[fpos] = now;
        fk = 'k';
    }
    else {
        ifd_mod[fpos] = now;
        if(ifd_ref[fpos]<0) {
            fk = 'z';
        }
        else {
            fk = '-';
        }
    }

    *rfd = -1;

#ifdef TRACE_FD
    Trace("ifd_close: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d; use %5d %c\n",
        e->h1, e->h2, e->h3, e->h4, fpos, xrfd, ifd_ref[fpos], nfd_cached, fk);
#endif

    MUTEX_UNLOCK(&fd_assign_lock);

out:
    return ret;
}


#ifdef NP

int
bfd_close(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    return bfd_raw_close(rfd, bk, e);
}

#else


int
bfd_close(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    int xrfd;
    int chk;
    char fk;
    int now;

    ret = 0;

    xrfd = *rfd;
    if(*rfd<0) {
        goto out;
    }

    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = -5; 
    xe->h4 = -1; 

    fpos = FDPOOL_POS_E(xe);

    if(!fdcaching_flag) {
        ret = bfd_raw_close(rfd, bk, e);
        *rfd = -1;
        goto out;
    }

    MUTEX_LOCK(&fd_assign_lock);

    if(bfd_ref[fpos]>0) {
        bfd_ref[fpos]--;
        bfd_mod[fpos] = now;
        fk = 'k';
    }
    else {
        bfd_mod[fpos] = now;
        if(bfd_ref[fpos]<0) {
            fk = 'z';
        }
        else {
            fk = '-';
        }
    }

    *rfd = -1;

#ifdef TRACE_FD
    Trace("bfd_close: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d; use %5d %c\n",
        e->h1, e->h2, e->h3, e->h4, fpos, xrfd, bfd_ref[fpos], nfd_cached, fk);
#endif

    MUTEX_UNLOCK(&fd_assign_lock);

out:
    return ret;
}

#endif

#if 0
inline
static
#endif
int
fd_compaction()
{
    int i;
    int fdpoolsize = radix_h1*radix_h2;
    int now;
    int sp, ep, j;
    int ip, bp;
    int ac, aci, acb;
    int aim;
    int lv;
    struct timeval zbefore, zafter;


    MUTEX_LOCK(&fd_assign_lock);
    j = fd_comp_running;
    if(j==0) {
        fd_comp_running++;
    }
    lv  = nfd_cached; 
    sp  = (fd_comp_last + 1)%fdpoolsize;
    /*
    aim = nfd_cached - nfd_cached_lwater;
    */
    MUTEX_UNLOCK(&fd_assign_lock);
    aim = fd_comp_width;

    if(j) {
        return 0;
    }


    now = (int)time(NULL);
#if 0
    for(i=0;i<fdpoolsize;i++) {
#if 0
        Trace(" %4d: ifd %4d, mod %d (diff %5d), ref %3d\n",
            i, ifd_map[i], ifd_mod[i], now - ifd_mod[i], ifd_ref[i]);
#endif
        if(ifd_map[i]>=0) {
        Trace("pos %6d; ifd %4d, mod %d (diff %5d), ref %3d\n",
            i, ifd_map[i], ifd_mod[i], now - ifd_mod[i], ifd_ref[i]);
        }
    }
#endif

    gettimeofday(&zbefore, NULL);


    ac = aci = acb = 0;
#if 0
    for(j=0; nfd_cached>nfd_cached_lwater && j<fdpoolsize;j++)
#endif
    for(j=0; ac<aim && nfd_cached>nfd_cached_lwater && j<fdpoolsize;j++)
    {
        i = (sp+j)%fdpoolsize;
        ip = bp = 0;

        MUTEX_LOCK(&fd_assign_lock);
        if(ifd_map[i]>=0 && ifd_ref[i]<=0 && now - ifd_mod[i] > ifd_keeptime) {
            ip = 1;
        }
        if(bfd_map[i]>=0 && bfd_ref[i]<=0 && now - bfd_mod[i] > bfd_keeptime) {
            bp = 1;
        }
        MUTEX_UNLOCK(&fd_assign_lock);

        if(ip) {
            ifd_raw_closeP(i);
            aci++;
        }
        if(bp) {
            bfd_raw_closeP(i);
            acb++;
        }
        ac = aci + acb;
        ep = i;
    }

    gettimeofday(&zafter, NULL);

#if 0
    Trace(
      "fd_comp: %4d scan %5d %05d-%05d; close %5d %5d %5d; hold %5d-> %5d\n",
        KOTETU_SUB_MSEC(zafter, zbefore),
        (sp>ep ? ep+fdpoolsize-sp : ep-sp), sp, ep,
        ac, aci, acb, lv, nfd_cached);
#endif

    Trace(
      "fd_comp: %5d %4d; s %5d %05d-%05d; c %4d %4d %4d; h %5d-> %5d\n",
        now - boot_date, KOTETU_SUB_MSEC(zafter, zbefore),
        (sp>ep ? ep+fdpoolsize-sp : ep-sp), sp, ep,
        ac, aci, acb, lv, nfd_cached);

    if(ac==0) {
        Error("fd_comp:  no fd slot");
    }

    MUTEX_LOCK(&fd_assign_lock);
    fd_comp_running = 0;
    fd_comp_last = ep;
    MUTEX_UNLOCK(&fd_assign_lock);

    return ac;
}

int
ifd_compaction()
{
    return fd_compaction();
}

int
bfd_compaction()
{
    return fd_compaction();
}

/* ctype
        1   -   check equal or not
        2   -   check empty of not
*/
int
CheckFileSize(int ctype, char *idmsg, char *name, off_t exp, off_t esz)
{
    struct stat sbuf;
    int chk;

#ifndef FSIZE_CHECK
    return 0;
#endif

    chk = stat(name, &sbuf);
    if(chk) {
        Error("fail stat() for %s '%s' (%d)", idmsg, name, errno);
        return -1;
    }

    if(ctype==1) {
        if(sbuf.st_size != (off_t) exp) {
            Trace("%s %-20s ignore-size %9lldB %5lds ; expect %9lldB %6ldB x %lds\n",
                idmsg, name, 
                (long long)sbuf.st_size, (long)((long long)sbuf.st_size/esz),
                (long long) exp, (long)esz, (long)(exp/esz));
            return 1;
        }
        else {
            return 0;
        }
    }
    else
    if(ctype==2) {
        if(sbuf.st_size == (off_t) exp) {
            return 0;
        }
        if(sbuf.st_size != (off_t) 0) {
            Trace("%s %-20s not-empty %9lldB %5lds ; expect %9lldB %6ldB x %lds\n",
                idmsg, name, 
                (long long)sbuf.st_size, (long)((long long)sbuf.st_size/esz),
                (long long) exp, (long)esz, (long)(exp/esz));
            return 1;
        }
        else {
            return 0;
        }
    }
    else {
            Trace("%s %-20s actual %9lldB %5lds ; expect %9lldB %6ldB x %lds\n",
                idmsg, name, 
                (long long)sbuf.st_size, (long)((long long)sbuf.st_size/esz),
                (long long) exp, (long)esz, (long)(exp/esz));
        return 0;
    }

    return 0;
}


/*
 * return:
 *      0   success
 *      !0  fail and indicate reason
 *
 */
int
ifd_raw_open(int *rfd, bucket_t *bk, skcache_entry_t *e, int szchk)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;
    int now;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif
    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = -1; 
    xe->h4 = -1; 

#if 0
    fpos = FDPOOL_POS_E(xe);
#endif

    fpos = FDPOOL_POS(e->h1, e->h2);

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)Oblksize,
        (long)fpos);
#endif


/*
    namechk = assign_indexfilename(name, xe);
*/
    namechk = assign_indexfilename(name, e);

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] path '%s'\n",
        e->h1, e->h2, e->h3, e->h4, name);
#endif

    c = 0;
try_open:
    c++;
    CX(iopen);
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2);
            goto try_open;
        }
        if(errno==ENOSPC) {
            Error("ifd_open: disk full when open file '%s'", name);
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("ifd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = XSK_INDEX_EOPEN;
        goto fail_open;
    }

    if(szchk) {
        CheckFileSize(szchk, "INDEX", name,
            (off_t) Onblk_file*ESIZE, ESIZE);
    }

    MUTEX_LOCK(&fd_assign_lock);
    ifd_map[fpos] = sfd;
    ifd_mod[fpos] = now;
    ifd_ref[fpos]++;
    nfd_cached++;
    MUTEX_UNLOCK(&fd_assign_lock);

    *rfd = sfd;

#ifdef TRACE_POS2
Trace("ifd_open: new    %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;

out:


#ifdef TRACE_POS2
Trace("ifd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

#ifdef TRACE_FD
    Trace("ifd_raw_o: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d, use %5d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, *rfd, ifd_ref[fpos], nfd_cached);
#endif

    return ret;
}


int
bfd_raw_open(int *rfd, bucket_t *bk, skcache_entry_t *e, int szchk)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;
    int now;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS2
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif

    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
#ifdef PP
    xe->h3 = -7; 
    xe->h4 = -1; 
#endif

#if 0
    fpos = FDPOOL_POS_E(xe);
#endif
    fpos = FDPOOL_POS(e->h1, e->h2);

#ifdef TRACE_POS2
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)Oblksize,
        (long)fpos);
#endif

    namechk = assign_bodyfilename(name, e);

    c = 0;
try_open:
    c++;
    CX(popen);
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2);
            goto try_open;
        }
        if(errno==ENOSPC) {
            Error("bfd_open: disk full when open file '%s'", name);
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("bfd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = XSK_BODY_EOPEN;
        goto fail_open;
    }

    if(szchk) {
        CheckFileSize(szchk, "BODY ", name,
            (off_t) Onblk_file*Oblksize, Oblksize);
    }

    MUTEX_LOCK(&fd_assign_lock);
    bfd_map[fpos] = sfd;
    bfd_mod[fpos] = now;
    bfd_ref[fpos]++;
    nfd_cached++;
    MUTEX_UNLOCK(&fd_assign_lock);

    *rfd = sfd;

#ifdef TRACE_POS2
Trace("bfd_open: new    %d,%d sfd %d, *rfd %d, return %d\n",
    xe->h1, xe->h2, sfd, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;

out:

#ifdef TRACE_POS2
Trace("bfd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

#ifdef TRACE_FD
    Trace("bfd_raw_o: [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d; use %5d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, *rfd, bfd_ref[fpos], nfd_cached);
#endif

    return ret;
}

/*
 * return:
 *      0   success
 *      !0  fail and indicate reason
 *
 */
int
ifd_open(int *rfd, bucket_t *bk, skcache_entry_t *e, int szchk)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;
    int now;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif
    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = -1; 
    xe->h4 = -1; 

    fpos = FDPOOL_POS_E(xe);

    if(!fdcaching_flag) {
        ret = ifd_raw_open(rfd, bk, e, szchk);
        goto out;
    }

#ifdef TRACE_POS2
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)Oblksize,
        (long)fpos);
#endif

    MUTEX_LOCK(&fd_assign_lock);
    sfd = ifd_map[fpos];
    MUTEX_UNLOCK(&fd_assign_lock);

    if(sfd>=0) {
        *rfd = sfd;
#ifdef TRACE_POS2
Trace("ifd_open: cached %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif
        MUTEX_LOCK(&fd_assign_lock);
        ifd_ref[fpos]++;
        ifd_mod[fpos] = now;
        MUTEX_UNLOCK(&fd_assign_lock);

        ret = 0;
    }
    else {
        if(nfd_cached > nfd_cached_hhwater) {
            ifd_compaction();
        }
#if 0
        if(nfd_cached > nfd_cached_hwater) {
            ifd_compaction();
        }
#endif

        ret = ifd_raw_open(rfd, bk, e, szchk);
    }
out:

#ifdef TRACE_POS2
Trace("ifd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

#ifdef TRACE_FD
    Trace("ifd_open:  [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d, use %5d ret %d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, *rfd, ifd_ref[fpos], nfd_cached, ret);
#endif

    return ret;
}

#ifdef NP

int
bfd_open(int *rfd, bucket_t *bk, skcache_entry_t *e, int szchk)
{
    return bfd_raw_open(rfd, bk, e, szchk);
}

#else


int
bfd_open(int *rfd, bucket_t *bk, skcache_entry_t *e, int szchk)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;
    struct stat sbuf;
    int chk;
    int now;

    ret = 0;
    *rfd = -1;

#ifdef TRACE_POS2
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif
    now = (int)time(NULL);

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = -9; 
    xe->h4 = -1; 

    fpos = FDPOOL_POS_E(xe);

    if(!fdcaching_flag) {
        ret = bfd_raw_open(rfd, bk, e, szchk);
        goto out;
    }

#ifdef TRACE_POS2
    Trace("bfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)Oblksize,
        (long)fpos);
#endif

    MUTEX_LOCK(&fd_assign_lock);
    sfd = bfd_map[fpos];
    MUTEX_UNLOCK(&fd_assign_lock);

    if(sfd>=0) {
        *rfd = sfd;
#ifdef TRACE_POS2
Trace("bfd_open: cached %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif
        MUTEX_LOCK(&fd_assign_lock);
        bfd_ref[fpos]++;
        bfd_mod[fpos] = now;
        MUTEX_UNLOCK(&fd_assign_lock);

        ret = 0;
    }
    else {
        if(nfd_cached > nfd_cached_hhwater) {
            bfd_compaction();
        }
#if 0
        if(nfd_cached > nfd_cached_hwater) {
            bfd_compaction();
        }
#endif

        ret = bfd_raw_open(rfd, bk, e, szchk);
    }
out:

#ifdef TRACE_POS2
Trace("bfd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

#ifdef TRACE_FD
    Trace("bfd_open:  [%03d:%03d:%03d:%04d] pos %6d; fd# %5d %2d, use %5d ret %d\n",
        e->h1, e->h2, e->h3, e->h4, fpos, *rfd, bfd_ref[fpos], nfd_cached, ret);
#endif

    return ret;
}

#endif



int
bfd_delete(int *rfd, bucket_t *bk, skcache_entry_t *e, int notify)
{
    int ret;
    char *namechk;
    char name[L_PATH];
    int jfd;
    int chk;

    ret = 0;
    *rfd = -1;

#if 0
 {
    int fpos;
    fpos = FDPOOL_POS_E(e);
    Trace("bfd_delete: st dir# %d, h1,h2,h3,h4=%d,%d,%d,%d oid# %d\n",
        fpos, e->h1, e->h2, e->h3, e->h4, e->oid);
 }
#endif

#if 0
 {
    int fpos;
    fpos = FDPOOL_POS_E(e);
    Trace("bfd_delete: be dir# %d, oid# %d, bfd map %d, mod %d, ref %d\n",
        fpos, e->oid, bfd_map[fpos], bfd_mod[fpos], bfd_ref[fpos]);
 }
#endif

    jfd = 1234567890;

    namechk = assign_bodyfilename(name, e);
    chk = bfd_close(&jfd, bk, e);

    chk = unlink(name);
    if(chk<0) {
        if(errno==2) {
            if(notify) {
            Trace("bfd_delete: no-existance file '%s' (%d)\n", name, errno);
            }
            ret = -2;
        }
        else {
            Error("bfd_delete: unlink '%s' fail (%d)", name, errno);
            ret = -1;
        }
    }
#if 0
    else {
        Trace("bfd_delete: unlink '%s' success\n", name);
    }
#endif

#if 0
 {
    int fpos;
    fpos = FDPOOL_POS_E(e);
    Trace("bfd_delete: af dir# %d, oid# %d, bfd map %d, mod %d, ref %d\n",
        fpos, e->oid, bfd_map[fpos], bfd_mod[fpos], bfd_ref[fpos]);
 }
#endif

    return ret;
}



/*
 * clear packed file (insert ignore magic number).
 * additionaly, create if it is not exist.
 */

int
build_file(int d1, int d2)
{
    bucket_t *bk;
    skcache_entry_t tmp_ent;
    skcache_entry_t cur_ent;
    skcache_entry_t entbulk;
    char blkbulk[Oblksize];
    struct iovec iiov[HARD_NINDEX_BDWRITE];
    struct iovec biov[HARD_NINDEX_BDWRITE];
    int j;
    int chk;
    int ifd, bfd;
    int sig_pos = -1;
    int rest;
    int pos;
    off_t nw;
    off_t dst;

    entbulk.magic[0] = MAGIC_0 & 0x99;
    entbulk.magic[1] = MAGIC_1 & 0x66;
    memset(blkbulk, '=', Oblksize);

    for(j=0;j<nindex_bdwrite;j++) {
        iiov[j].iov_base = &entbulk;
        iiov[j].iov_len  = ESIZE;
        biov[j].iov_base = blkbulk;
        biov[j].iov_len  = Oblksize;
    }

    sig_pos = DSIG_POS(d1, d2, 0);
    bk = &skcache_bucket_base[d1];

    tmp_ent.h1 = d1;
    tmp_ent.h2 = d2;
    tmp_ent.h3 = -3;
    tmp_ent.h4 = -1;

    chk = ifd_open(&ifd, bk, &tmp_ent, 2);
    if(chk) {
        Error("build_file: fail ifd_open() for [%d:%d:%d:%d]",
            tmp_ent.h1, tmp_ent.h2, tmp_ent.h3, tmp_ent.h4);
        goto fail_open;
    }
    if(ifd<0) {
        Error("build_file: ignore fd# of index file for [%d:%d:%d:%d]",
            tmp_ent.h1, tmp_ent.h2, tmp_ent.h3, tmp_ent.h4);
        goto fail_open;
    }

#ifdef PP
    chk = bfd_open(&bfd, bk, &tmp_ent, 2);
    if(chk) {
        Error("build_file: fail bfd_open() for [%d:%d:%d:%d]",
            tmp_ent.h1, tmp_ent.h2, tmp_ent.h3, tmp_ent.h4);
        goto fail_open2;
    }
    if(ifd<0) {
        Error("build_file: ignore fd# of body file for [%d:%d:%d:%d]",
            tmp_ent.h1, tmp_ent.h2, tmp_ent.h3, tmp_ent.h4);
        goto fail_open2;
    }
#endif

    dst = Iseek(sig_pos, ifd, (off_t)0);
#ifdef PP
    dst = Pseek(sig_pos, bfd, (off_t)0);
#endif

    pos = 0;
    do {
        rest = Onblk_file - pos;
        if(rest > nindex_bdwrite) {
            rest = nindex_bdwrite;
        }
        if(rest<=0) {
            break;
        }

#if 0
        Trace(" file %3d:%02d pos %d/%d, rest %d\n",
            d1, d2, pos, Onblk_file, rest);
#endif
        nw = Iwritev(sig_pos, ifd, iiov, rest);
#ifdef PP
        nw = Pwritev(sig_pos, bfd, biov, rest);
#endif
        pos += rest;

    } while(rest>0);

#ifdef PP
    chk = bfd_close(&bfd, bk, &tmp_ent);
#endif
    chk = ifd_close(&ifd, bk, &tmp_ent);

    goto out;

fail_open2:
    chk = ifd_close(&ifd, bk, &tmp_ent);

fail_open:

    return -1;

out:

    return 0;
}


void
dbit_mkhumanstr(u_int32_t *s, char *line)
{
    int i;
    u_int32_t m, v;
    char *p;

    p = line;
    for(i=0;i<DBIT_BIT_WIDTH;i++) {
        m = 1U << i%(SIZE_OF_UINT*8);
        v = s[i/(SIZE_OF_UINT*8)] & m;
        *p++ = (v ? '1' : '0');
    }
    *p = '\0';
}

void
dbit_dump(u_int32_t *s)
{
    char line[DBIT_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */

    dbit_mkhumanstr(s, line);
    Trace("%s\n", line);
}

void
dbit_show(u_int32_t *s)
{
    char line[DBIT_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */
    int i, c;

    dbit_mkhumanstr(s, line);
    c = 0;
    for(i=0;i<DBIT_BIT_WIDTH;i++) {
        if(DBIT_TMP_ISSET(s, i)) {
            c++;
        }
    }
    Trace("addr %p count %d, bits %s\n", s, c, line);
}


void
dbit_set(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    s[i/(SIZE_OF_UINT*8)] |= m;
}

int
dbit_isset(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    return (s[i/(SIZE_OF_UINT*8)] & m) ? 1 : 0;
}

int
dbit_allset(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DBIT_ELEM_WIDTH;i++) {
        u = (u_int32_t)-1L; /* all bits were set */
        s[i] = u;
    }
}

int
dbit_allclear(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DBIT_ELEM_WIDTH;i++) {
        u = (u_int32_t)0UL;
        s[i] = u;
    }
}










/*
#define ofile_offset(e) ((e->h3+e->h4)%Onblk_file)
*/

inline
int
old_ofile_offset(skcache_entry_t *e)
{
#if 0
    return (e->h3+e->h4)%Onblk_file;
#endif
    return ((e->h3*e->h4)%(Onblk_file/Onblk_obj))*Onblk_obj;
}

/*
 it is defined in ocache.h
#define ofile_offset(e) (((e->h3*e->h4)%(Onblk_file/Onblk_obj))*Onblk_obj)
*/


#if 0

int
dump_ifile(int ifd, int start_pos)
{
    int i;
    int n;
    skcache_entry_t tmp_entry;
    off_t pos;
    int nr;
    int diff;

    n = start_pos;
    for(i=0;i<Onblk_seek;i++) {
        pos = n*ESIZE;
        lseek(ifd, pos, SEEK_SET);

        nr = read(ifd, &tmp_entry, ESIZE);

        diff = tmp_entry.issue_time - min_date;
    Trace("  :: fd# %d, try %3d: pos %4d %12lldB : magic(hex) %02x %02x, oid# %d, [%d] diff %d\n",
            ifd, i, n, (long long)pos,
            (unsigned char)tmp_entry.magic[0],
            (unsigned char)tmp_entry.magic[1],
            tmp_entry.oid, tmp_entry.h4, diff
        );

        n = (n+1)%Onblk_file;
    }

    return 0;
}

#endif



dump_syscalls(FILE *fp)
{
    char *fmt = "%-32s %32s\n";
    
    fprintf(fp, fmt, "name", "value");
    fprintf(fp, fmt, "====", "=====");

    SH(fp, iopen);
    SH(fp, iclose);
    SH(fp, ilseek);
    SH(fp, iread);
    SH(fp, ireadv);
    SH(fp, iwrite);
    SH(fp, iwritev);

    SH(fp, popen);
    SH(fp, pclose);
    SH(fp, plseek);
    SH(fp, pread);
    SH(fp, preadv);
    SH(fp, pwrite);
    SH(fp, pwritev);

    SH(fp, junk);

}


/* MASK system calls */

#define read(x1,x2,x3)  "error" __FILE__ __LINE__
#define write(x1,x2,x3) "error" __FILE__ __LINE__
#define readv(x1,x2,x3) "error" __FILE__ __LINE__
#define writev(x1,x2,x3)    "error" __FILE__ __LINE__

/***
 *** end of low I/O
 ***/

