/* 
 * reaper3.c :
 *      memory sweeper
 *      memory garbage collection routine for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2001.
 *
 */

char* skcache_label_long(void *pp);


static
void
show_targets(time_t reftime, skcache_gc_target_t *tbl, int n, int bid)
{
    int o;

    for(o=0;o<n;o++) {
        if(tbl[o].pos) {
        Trace(
" %3d: b %3d bk# %d, oid# %d, pos %p, %d:%d:%d:%d count %d, hot %4d\n",
            o, tbl[o].no, bid, tbl[o].oid, tbl[o].pos,
            tbl[o].pos->h1, tbl[o].pos->h2, tbl[o].pos->h3, tbl[o].pos->h4,
            tbl[o].pos->count,
            reftime - tbl[o].hot);
        }
        else {
        Trace(" %3d: b %3d bk# %d, oid# %d, pos %p, -:-:-:- hot %4d\n",
            o, tbl[o].no, bid, tbl[o].oid, tbl[o].pos, reftime - tbl[o].hot);
        }
    }
}



inline
int
comp_targets_hot(const void *xa, const void *xb)
{
    skcache_gc_target_t *a = (skcache_gc_target_t*) xa;
    skcache_gc_target_t *b = (skcache_gc_target_t*) xb;

    if(a==NULL || a->oid==0) {
        return 1;
    }
    if(b==NULL|| b->oid==0) {
        return -1;
    }
    return a->hot - b->hot;
}

/*
int ofile_offset(skcache_entry_t *e);
*/

inline
int
comp_targets_pos(const void *xa, const void *xb)
{
    skcache_gc_target_t *a = (skcache_gc_target_t*) xa;
    skcache_gc_target_t *b = (skcache_gc_target_t*) xb;
    int av, bv;

    if(a==NULL || a->oid==0 || a->pos==NULL) {
        return 1;
    }
    if(b==NULL|| b->oid==0 || b->pos==NULL) {
        return -1;
    }
    av = ofile_offset(a->pos);
    bv = ofile_offset(b->pos);

    return av - bv;
}

#ifndef TMP_NTARGET
#define TMP_NTARGET     (2048)
#endif
#ifndef SWAPOUT_CHUNK_TIMEOUT
#define SWAPOUT_CHUNK_TIMEOUT   (3000)  /* 3 sec */
#endif

int noslot_count_limit=3;

int ntry_swapout = 3;

/*
 * reaper3
 */
int
reaper(int dnum, bucket_t *bk, int nq, int force, int verbose, int opt)
{
    int ret;
    int chk;
    int csum;
    int action;
    int ntry;
    int new_keep, last_keep;
    int i;
    int cp;
    long long bsum;
    int act_sum[6];
    skiplist_t *list;
    skiplist_cell_t *x, *n;
    skiplist_cell_t *last;
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skcache_entry_t *ent;
    struct timeval xbefore, xafter;
    struct timeval xmiddle, xlap;
    struct timeval sobegin, soend;
    time_t now;
    int checkpoint;
    int target_hot;
    skcache_entry_t *target_pos;
    int tmp_ntarget;
    skcache_gc_target_t tmp_targets[TMP_NTARGET];
    int c;
    int nf;
    int max_cp;
    int wait_flag;
    int sp_ntarget, last_ntarget;
    int fpos;

    extern time_t boot_date;


    /* NOTE:
     *  bucket is already locked. don't re-lock.
     */

    gettimeofday(&xbefore, NULL);
    now = xbefore.tv_sec;

    FCHK();

    sp_ntarget = nq;
    if(nq<0) {
        nq = MAX_GC_NTARGET;
    }
    last_ntarget = bk->gc_ntarget;

    MUTEX_LOCK(&stat_lock);
    new_keep = skcache_mem_keep_cur;
    MUTEX_UNLOCK(&stat_lock);


#ifdef TRACE_REAPER
#endif
if(verbose)
    Trace(
"reaper3: S t %d, th# %d, bk# %d, nq %d (%d), time %d, target %d, elem %4d/%d, entry %d+%d %d, frag %d\n",
        xbefore.tv_sec, 
        opt,
        bk->bid, nq, sp_ntarget,
        bk->gc_last,
        bk->gc_ntarget,
        bk->cup.num, bk->elemnum,
        bk->entry_use, bk->entry_free, bk->entry_all,
        bk->own_frag);


assert(bk->cup.num==bk->entry_use);

    ret = 0;

    for(i=0;i<6;i++) {
        act_sum[i] = 0;
    }

    target_hot = now;
    target_pos = NULL;

    if(bk->gc_ntarget>0 && xbefore.tv_sec < bk->gc_last) {
#ifdef TRACE_REAPER
        if(verbose)
        Trace("  skip scanning bk# %d, ntarget %d (hot %d sec)\n",
            bk->bid, bk->gc_ntarget, bk->gc_last - xbefore.tv_sec);
#endif
        goto delete_targets;
    }


    /***
     ***
     ***/

    MUTEX_LOCK(&stat_lock);
    msweeper_count++;
    MUTEX_UNLOCK(&stat_lock);


    list = &bk->cup;


    if(0*verbose) {
        Trace("before\n");
        bucket_show(bk);
    }


#if 1
    /* clear headers */
    list->level = 0;
    x = list->header;
    x->entity = list->term;
    for(i=1;i<=list->maxlevel;i++) {
        x->skip[i] = list->term;
    }
#endif


    ntry = 0;

try_sweep:
    ntry++;

    for(i=0;i<6;i++) {
        act_sum[i] = 0;
    }

assert(bk->cup.num==bk->entry_use);
#ifdef TRACE_REAPER
#endif
    if(verbose)
    Trace("  bk# %d, try %d, entries %d (%d,%d), keep %d sec, global %d sec \n",
        bk->bid, ntry, bk->entry_use, bk->entry_hwater, bk->entry_lwater,
        new_keep, skcache_mem_keep_cur);

    csum = 0;
    bsum = 0LL;
    cp = 0;

    max_cp = bk->entry_use;

    x = list->header;
    x = x->skip[0];
    last = NULL;

    tmp_ntarget = 0;

    while(x!=list->term) {
assert(bk->cup.num==bk->entry_use);

        if(cp>=max_cp) {
            Error("cp %d is too big than max_cp %d", cp, max_cp);
        }

        cp++;
        n = x->skip[0];
        ent = x->entity;
#ifndef NDEBUG
        if(ent==NULL) {
            Error("bk# %d, %p/ %d th %x has empty elem.",
                bk->bid, bk, cp);
            goto next_elem;
        }
#endif
#ifdef CHECK_MAGIC
        if(ent->magic[0]!=MAGIC_0|| ent->magic[1]!=MAGIC_1) {
            Error("reaper3: ignore magic bk# %d, ent %p", bk->bid, ent);
            goto next_elem;
        }
#endif

#if 1
        for(i=1;i<=list->maxlevel;i++) {
            x->skip[i] = list->term;
        }
#endif

#if 0
        if(verbose && ent->ref>0 ) {
            Trace(" ent %p; ref %d\n", ent, ent->ref );
        }
#endif

#if 0
        action = skcache_dicision_w_th(bk, ent, 0,
            xbefore.tv_sec,
            skcache_mem_keep_min, new_keep,
            skcache_entry_keep_cur);
#endif
        action = skcache_dicision_w_th(bk, ent, 0,
            xbefore.tv_sec,
            3, 3, 4);

#if 0
        if(action==2) {
            action = 3;
        }
#endif

        act_sum[action]++;

#if 0
 Trace("  %3d: %p: entity %#x action %d\n", cp, x, ent, action);
#endif
        switch(action) {
        case 0: /* KEEP */
            csum++;
            bsum += ent->totallength;

            if(ent->hot>checkpoint) {
                checkpoint = ent->hot;
            }

            if(ent->hot<target_hot) {
                target_hot = ent->hot;
                target_pos = ent;
            }

            break;

        case 1: /* SWAPOUT */
        case 3: /* SWAPOUT + PURGE */
            /* mark */

            if(tmp_ntarget<TMP_NTARGET) {
#if 0
Trace("store  bk# %d, %d/%-2d : pos %p oid# %d\n",
    bk->bid, tmp_ntarget, TMP_NTARGET,  ent, ent->oid);
#endif
                tmp_targets[tmp_ntarget].no  = cp;
                tmp_targets[tmp_ntarget].pos = ent;
                tmp_targets[tmp_ntarget].hot = ent->hot;
                tmp_targets[tmp_ntarget].oid = ent->oid;
                tmp_ntarget++;
            }


            break;

        case 2: /* PURGE */
purge_elem:
            /* frag level */
            skcache_entry_purge(bk, ent);

            /* entry level */
            skcache_push_idle_entry(bk, ent);
            ent = NULL;
            x->entity = NULL;

#if 0
            /* elem level */
            for(i=1;i<=list->maxlevel;i++) {
                x->skip[i] = list->term;
            }
#endif

            /* release from skiplist */
            if(last==NULL) {
                list->header->skip[0] = n;
            }
            else{
                last->skip[0] = n;
            }

            x->skip[0] = bk->elem_freelist;
            bk->elem_freelist = x;
            list->num--;

#ifdef TRACE_REAPER
#ifndef NDEBUG
 {
    skiplist_cell_t *y;
    int m, k;

    y = list->header;
    y = y->skip[0];
    m = 0;
    while(y!=list->term) {
        if(y->entity==NULL) {
            Trace("re:ch: bk# %d, %3d th entity is empty\n",
                bk->bid, m);
        }
        for(k=0;k<=list->maxlevel;k++) {
            if(y->skip[k]==NULL) {
                Trace("re:ch: bk# %d, %3d th entry, %d th skip is NULL\n",
                    bk->bid, m, k);
            }
        }

        y = y->skip[0];
        m++;
    }
 }
#endif
#endif

            break;

        default:
            break;
        }

        if(action!=2) {
            last = x;
        }

next_elem:
        x = n;
    }

#if 0
  Trace("  use %d, count %d, keeptime %d, hold %d entries %lld bytes (%.2fMB)\n\
  action bk# %d, keep %d, swap %d, purge %d, swap+purge %d, markd %d, unlink %d\n",
        bk->entry_use, cp, new_keep,
        csum, bsum, fdM(bsum),
        bk->bid,
        act_sum[0], act_sum[1], act_sum[2], act_sum[3], act_sum[4], act_sum[5]);
#endif

#if 1
    if(cp!=(act_sum[0]+act_sum[1]+act_sum[2]+act_sum[3]+act_sum[4]+act_sum[5])) {
        Error("strange A bk# %d, (sum of action %d, cp %d)",
            bk->bid,
          act_sum[0]+act_sum[1]+act_sum[2]+act_sum[3]+act_sum[4]+act_sum[5], cp);
    }
#endif


    if(tmp_ntarget>0) {
#if 0
        Trace("  bk# %d, tmp_ntarget %d\n", bk->bid, tmp_ntarget);
#endif

        if(0*verbose) {
            Trace("raw 0:\n");
            show_targets(now, tmp_targets,
                MIN(MAX_GC_NTARGET, tmp_ntarget), bk->bid);
        }

        qsort(tmp_targets, tmp_ntarget,
            sizeof(skcache_gc_target_t), comp_targets_hot);

        if(0*verbose) {
            Trace("\n1\n");
            show_targets(now, tmp_targets,
                MIN(MAX_GC_NTARGET, tmp_ntarget), bk->bid);
        }

        bk->gc_rest = tmp_ntarget;
        bk->gc_ntarget = MIN(tmp_ntarget, MAX_GC_NTARGET);
        
#if 0
        for(i=0;i<MIN(tmp_ntarget, MAX_GC_NTARGET);i++) {
            bk->gc_targets[i] = tmp_targets[i];
        }
        bk->gc_ntarget = i;
#endif

#if 0
        for(i=0;i<bk->gc_ntarget;i++) {
            Trace("store  bk# %d, %d/%-2d : pos %p oid# %d\n",
                bk->bid, i, MAX_GC_NTARGET, 
                bk->gc_targets[i].pos, bk->gc_targets[i].oid);
        }
#endif

    }

#if 0

#if 1   /* cleaning still low water */
    else
    if(bk->entry_use <= bk->entry_lwater) {
#ifdef TRACE_REAPER
        if(verbose)
        Trace("  Okay<%d>, bk# %d, %d/%d\n",
            ntry, bk->bid, bk->entry_use, bk->entry_lwater);
#endif
        bk->gc_rest = 0;
        bk->gc_ntarget = 0;
    }
    else
    if(ntry<5) {
        last_keep = new_keep;
        new_keep = 6*new_keep/10;

#ifdef TRACE_REAPER
        if(verbose)
        Trace("  retry<%d>, bk# %d, %d/%d new keep %d\n",
            ntry, bk->bid, bk->entry_use, bk->entry_lwater, new_keep);
#endif
        goto try_sweep;
    }
    else{
#ifdef TRACE_REAPER
        if(verbose)
        Trace("  giveup<%d>, bk# %d, %d/%d\n",
            ntry, bk->bid, bk->entry_use, bk->entry_lwater);
#endif
#if 0
        bk->gc_rest = 0;
        bk->gc_ntarget = 0;
#endif
    }


#else
    else{
        bk->gc_rest = 0;
        bk->gc_ntarget = 0;
    }
#endif

#endif

    bk->gc_last = xbefore.tv_sec+5;



    gettimeofday(&xmiddle, NULL);


delete_targets:
 {
    int dc;
    int dc_min;
    int o;
    int w, wsum;
    int lap;
    int noslot_count;

    noslot_count = 0;

    dc = 0;
    if(bk->entry_use > bk->entry_lwater) {
        if(bk->entry_use - bk->gc_ntarget > bk->entry_lwater) {
            Error("  bk# %d, cannot remove until lwater", bk->bid);
        }
        dc = bk->entry_use - bk->entry_lwater;
        if(dc<0) {
            dc = 0;
        }
    }

    dc_min = 0;
    if(bk->entry_use > bk->entry_hwater) {
        if(bk->entry_use - bk->gc_ntarget > bk->entry_hwater) {
            Error("  bk# %d, cannot remove until hwatern", bk->bid);
        }
        dc_min = bk->entry_use - bk->entry_hwater;
        if(dc_min<0) {
            dc_min = 0;
        }
    }

    if(verbose) {
        Trace("  bk# %d, sweeped, entry use %d, gray %d, water %d/%d/%d -> dc %d %d\n",
            bk->bid, bk->entry_use, bk->gc_ntarget,
            bk->entry_hhwater, bk->entry_hwater, bk->entry_lwater,
            dc, dc_min);
    }
 
    c = 0;

    if(bk->gc_ntarget>0) {
#if 0
            Trace("reaper3: ntarget %d\n", bk->gc_ntarget);
#endif

        if(0*verbose) {
            Trace("\n3\n");
            show_targets(time(NULL), tmp_targets, dc, bk->bid);
        }

        qsort(tmp_targets, dc,
            sizeof(skcache_gc_target_t), comp_targets_pos);

        if(0*verbose) {
            Trace("choiced 4:\n");
            show_targets(time(NULL), tmp_targets, dc, bk->bid);
        }

        bk->gc_ntarget = 0;
        for(i=0;i<dc;i++) {
            bk->gc_targets[i] = tmp_targets[i];
            bk->gc_ntarget++;
        }

        wsum = 0;
        for(o=0;o<dc;o++) {
            target_pos = bk->gc_targets[o].pos;
            if(target_pos) {
                if(target_pos->oid!=0) {
                    fpos = FDPOOL_POS_E(target_pos);
#if 0
                    w = (target_pos->totallength+Oblksize-1)/Oblksize;
#endif
                    w = NBCEIL(target_pos->totallength, Oblksize);

                    if(target_pos->ref) {
                        Trace(" bk# %d, entry %p, ref %d\n", 
                            bk->bid, target_pos, target_pos->ref);
                    }

                    gettimeofday(&sobegin, NULL);
                    chk = skcache_entry_swapout(bk, target_pos);
                    gettimeofday(&soend, NULL);

                    if(verbose) {
                        Trace(
" :swapout t %d bk# %d, o %3d, p %p, oid# %d, hash %d:%d:%d:%d dir# %d, pos %ld, lack %d ; last %d, len %ld %dblk ; %d msec, return %d\n",   
                        soend.tv_sec - boot_date,
                        bk->bid, o, target_pos,
                        target_pos->oid,
                        target_pos->h1, target_pos->h2,
                        target_pos->h3, target_pos->h4,
                        fpos, 
                        (long)ofile_offset(target_pos),
                        dir_lack[fpos],
                        now - target_pos->hot,
                        (long)target_pos->totallength, w,
                        KOTETU_SUB_TIME(soend, sobegin),
                        chk);
                    }
                    if(chk==0) {
                        wsum += w;
                    }
                    if(chk==XSK_SWAP_NOTFOUND) {
                        noslot_count++;
#if 1
#if 1
            Trace("  1 XSK_SWAP_NOTFOUND oid# %d,\n", target_pos->oid);
#endif
                        chk = 0;
#endif
#if 0
                        if(target_pos->swapout_candidate>ntry_swapout) {
                            chk = 0;    /* too many wait.  delete */
                        }
                        target_pos->swapout_candidate++;
#endif
                    }
                }
                else {
                    chk = 0;
                }
                if(chk==0) {
                    chk = skcache_entry_delete(bk, target_pos);
                    if(chk) {
#if 1
                        Trace("fail entry_delete, chk %d\n", chk);
#endif
                    }
                    c++;
                    bk->gc_ntarget--;
                }
            }

            gettimeofday(&xlap, NULL);
            lap = KOTETU_SUB_TIME(xlap, xmiddle);
            if(lap > SWAPOUT_CHUNK_TIMEOUT) {
                if(verbose) {
                    Trace(
" = too long lock time %d/%d msec, during %d items, bk# %d, skip!\n",
                        lap, SWAPOUT_CHUNK_TIMEOUT, o, bk->bid);
                }
                break;
            }
#if 0
            if(noslot_count>noslot_count_limit) {
                Trace(
" :skip this sweep dir# %d. because noslot event occures %d times\n",
                    fpos, noslot_count, noslot_count_limit);
                break;
            }
#endif
            if(c>=nq) {
                if(verbose) {
                    Trace(
"  bk# %d, reach to limitation of one batch(nq) %d, proccessed(c) %d\n",
                        bk->bid, nq, c);
                }
                break;
            }
        }
#if 1
        /* reset target counting */
        bk->gc_ntarget = 0;
#endif

        if(verbose) {
            Trace(" ,, success %d/%d = %.2f%%, noslot %d times\n",
                c, dc, 100.0*c/dc, noslot_count);
        }

        MUTEX_LOCK(&stat_lock);
        compaction_count += c;
        MUTEX_UNLOCK(&stat_lock);

        /*
        bk->gc_checkpoint = bk->gc_targets[0].hot;
        */
        bk->gc_checkpoint = target_hot;

#ifdef TRACE_REAPER
        if(verbose)
        Trace("  bk# %d, LRU remove %d/%d, rest %d\n",
            bk->bid, c, nq, bk->gc_ntarget);
#endif
    }
    else
    if(target_pos) {
        chk = skcache_entry_swapout(bk, target_pos);

        if(chk==XSK_SWAP_NOTFOUND) {
            noslot_count++;
#if 1
            Trace("  2 XSK_SWAP_NOTFOUND oid# %d,\n", target_pos->oid);
#endif
        }


        if(chk<0) {
#if 0
            Trace("    skip reason %d\n", chk);
#endif
        }
        else {
 if(chk) {
    Error(" 2 fail swapout ent %p, ret %d", target_pos, chk);
 }
            target_pos->ref = 0; /* force */
            chk = skcache_entry_delete(bk, target_pos);
 if(chk) {
    Error(" 2 fail delete ent %p, ret %d", target_pos, chk);
 }
            target_pos = NULL;
            c++;
        }

        MUTEX_LOCK(&stat_lock);
        compaction_count++;
        MUTEX_UNLOCK(&stat_lock);

        bk->gc_checkpoint = target_hot;
        target_hot = -1;

#ifdef TRACE_REAPER
        if(verbose)
        Trace("  bk# %d, NRU remove %d/%d, rest %d\n",
            bk->bid, c, nq, bk->gc_ntarget);
#endif
    }
    else {
        bk->gc_checkpoint = -1;

#if 0
        Trace("  bk# %d, none\n",
            bk->bid);
#endif
    }

#if 0
    Trace("  bk# %d, remove %d/%d, rest %d\n",
        bk->bid, c, nq, bk->gc_ntarget);
#endif

    if(verbose) {
        gettimeofday(&xlap, NULL);
        lap = KOTETU_SUB_TIME(xlap, xmiddle);
        Trace(" bk# %d, block %d, takes %d msec, ave %.2f msec/block\n",
            bk->bid, wsum, lap, (double)1.0*lap/wsum);
        Trace(" bk# %d, object %d, takes %d msec, ave %.2f msec/object\n",
            bk->bid, c, lap, (double)1.0*lap/c);
    }
 }

send_mark:

out:
    gettimeofday(&xafter, NULL);
    sim_stat_add_value(&msweep_time, KOTETU_SUB_TIME(xafter, xbefore));
    sim_stat_add_value(&msweep_time_short, KOTETU_SUB_TIME(xafter, xbefore));

if(verbose)
    Trace(
"reaper3: E t %d, th# %d, bk# %d,; ret %d; target spec. %d(%d), last %d done %d, left %d; entry %d %d/%d; takes scan,del,total %d %d %d msec del-ave %.2f; act %d %d %d %d %d %d\n",
        xafter.tv_sec, opt, bk->bid, ret,
        nq, sp_ntarget, last_ntarget, c, bk->gc_ntarget,
        bk->entry_use, bk->entry_hhwater, bk->entry_hwater,
        KOTETU_SUB_TIME(xmiddle, xbefore),
        KOTETU_SUB_TIME(xafter, xmiddle),
        KOTETU_SUB_TIME(xafter, xbefore),
        (double)KOTETU_SUB_TIME(xafter, xmiddle)/c,
        act_sum[0], act_sum[1], act_sum[2], act_sum[3], act_sum[4], act_sum[5]
    );
#ifdef TRACE_REAPER
    if(verbose)
    Trace("  bk# %d, checkpoint %d <%d sec ago>, takes %d msec\n",
        bk->bid, bk->gc_checkpoint, now - bk->gc_checkpoint,
        KOTETU_SUB_TIME(xafter, xbefore));

#endif

    if(0*verbose) {
        Trace("after\n");
        bucket_show(bk);
    }

    FCHK();

    return ret;
}

